/*
 *  K6300 C library
 *  Author:  Pierre Bureau for K-Team SA
 *  Version: 1.0
 *  Date:    19.11.2002
 */

/*  This file is part of K6300 Image Grabber.
 *  (c) 2002 Pierre Bureau for K-Team SA <pbureau@k-team.com>
 *
 *  K6300 Image Grabber is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 *  K6300 Image Grabber is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with K6300 Image Grabber; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "kt_k6300.h"
#include "kt_serial.h"

const char* raw_image_f_1 = "Raw image fmt=1";
const char* raw_image_f_2 = "Raw image fmt=2";

static int full_range = 0;

/*******************************************************************/
/* Get a K6300 Image from the serial line, the image format string
 * must be retrieved using kt_command() and passed to this function.
 */

int k6300_start_download(K6300_Image * image,const char * format)
{
    image->wb_b=1.0;
    image->wb_r=1.0;

    if(strncmp(format,raw_image_f_1,XANSLEN) == 0)
    {
        image->size= 160*120;
        image->dx= 160;
        image->dy= 120;
        image->format= binary;
        image->download= 0;
        image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));
        return 0;

    } else {

    if(strncmp(format,raw_image_f_2,XANSLEN) == 0)
    {
        image->size= 160*120*2;
        image->dx= 160;
        image->dy= 120;
        image->format= ascii;
        image->download= 0;
        image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));
        return 0;
    } else {

        image->format= wrong;
        return -1;
    }
    }
}

int k6300_download(K6300_Image *image)
{
    char * data_ptr;
    int status;

    if(image->format != binary && image->format != ascii)
        return -1;

    data_ptr= image->data;
    data_ptr+= image->download;

    status=kt_read_buffer(data_ptr, image->size-image->download);
    fprintf(stderr,"stat: %d (%u/%u)\n",status,image->download,image->size);

    image->download+= status;

    if(image->download < image->size)
        return 0;
    else
    {
        /* Invert the image */
        k6300_invert_image(image);
        return 1;
    }
}

#ifdef __WIN32
int k6300_get_image(HANDLE fd, K6300_Image * image,const char * format)
{
    int format_ok= 0;
    int charcount= 0;
    DWORD nbytes;
    unsigned char * data_ptr;
    unsigned char dump[1];

    if(strncmp(format,raw_image_f_1,XANSLEN) == 0)
    {
        image->size= 160*120;
        image->dx= 160;
        image->dy= 120;
        image->format= 1;
        image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));
        format_ok= 1;
    }

    if(strncmp(format,raw_image_f_2,XANSLEN) == 0)
    {
        image->size= 160*120*2;
        image->dx= 160;
        image->dy= 120;
        image->format= 0;
        image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));
        format_ok= 1;
    }

    image->wb_b=1.0;
    image->wb_r=1.0;

    if(format_ok)
    {
        /* read the '\n' that should be hanging around */
        if(!ReadFile(fd,(LPVOID)dump,1,&nbytes,NULL))
            return -2;

        /* Download the image */
        data_ptr= image->data;

        while(charcount < image->size)
        {
            if(!ReadFile(fd,(LPVOID)data_ptr,1,&nbytes,NULL))
                return -2;
            else
            {
                charcount++;
                data_ptr++;
            }
        }

        /* Invert the image */
        k6300_invert_image(image);

        return 0;
    }
    else
        return -1;
}
#else

int k6300_get_image(int fd, K6300_Image * image,const char * format)
{
    int format_ok= 0;
    int charcount= 0;
    int stat;
    unsigned char * data_ptr;
    unsigned char dump[1];

    if(strncmp(format,raw_image_f_1,XANSLEN) == 0)
    {
        image->size= 160*120;
        image->dx= 160;
        image->dy= 120;
        image->format= 1;
        image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));
        format_ok= 1;
    }

    if(strncmp(format,raw_image_f_2,XANSLEN) == 0)
    {
        image->size= 160*120*2;
        image->dx= 160;
        image->dy= 120;
        image->format= 0;
        image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));
        format_ok= 1;
    }

    image->wb_b=1.0;
    image->wb_r=1.0;

    if(format_ok)
    {
        /* read the '\n' that should be hanging around */
        stat= read(fd,dump,1);
        if(stat < 0)
            return -2;

        /* Download the image */
        data_ptr= image->data;

        while(charcount < image->size)
        {
            stat= read(fd,data_ptr,1);
            if(stat < 0)
                return -2;
            else
            {
                charcount++;
                data_ptr++;
            }
        }

        /* Invert the image */
        k6300_invert_image(image);

        return 0;
    }
    else
        return -1;
}
#endif

/*******************************************************************/
int k6300_create_black_image(K6300_Image * image)
{
    unsigned char * data_ptr;
    int charcount=0;

    image->size= 160*120;
    image->dx= 160;
    image->dy= 120;
    image->format= binary;
    image->data= (unsigned char*)calloc(image->size,sizeof(unsigned char));

    data_ptr= image->data;

    while(charcount < image->size)
    {
        data_ptr[charcount]=0;
        charcount++;
    }

    return 0;
}

/*******************************************************************/
int k6300_invert_image(K6300_Image * image)
{
    unsigned char * data_copy;
    int x,y;

    data_copy= (unsigned char*)calloc(image->size,sizeof(unsigned char));

    for(x=0;x<image->size;x++)
        data_copy[x]= image->data[x];


    for (y = 0; y < image->dy; y++) {
        for (x = 0; x < image->dx; x++) {
            image->data[x+y*image->dx]= data_copy[x+(image->dy-y)*image->dx];
        }
    }

    free(data_copy);

    return 0;
}

/*******************************************************************/
int k6300_ascii_to_binary(K6300_Image * image)
{
    unsigned char table[256];
    unsigned char* dst;
    char*          src;
    int i,t;

    dst = image->data;
    src = image->data;

    if (image->format == binary)
        return -1;
    else
    {
        for (t = 0; t < 256; t++) {
            table[t] = 0;
        }
        for (t = '0'; t <= '9'; t++) {
            table[t] = t-'0';
        }
        for (t = 'A'; t <= 'F'; t++) {
            table[t] = t-'A'+10;
        }
        for (t = 'a'; t <= 'f'; t++) {
            table[t] = t-'f'+10;
        }
        for (i=0; i < 160*120; i++) {
            dst[0] = table[src[0]]*16 + table[src[1]];
            dst += 1;
            src += 2;
        }
        image->size= image->size/2;
    }

    return 0;
}

/*******************************************************************/
/* Convert a K6300 image to RGB data.
 */
int k6300_RGB_convert(K6300_Image * image, RGB_Image * dest,
                      float wr, float wg, float wb)
{
    int stat[256] = { 0 };
    int stat_max = 0;
    int stat_min = 255;
    int x,y,i;
    double range;

    unsigned char ** data_copy;
    unsigned char r[4];
    unsigned char g[8];
    unsigned char b[4];


    int b0,r0,g0,g1;
    int yy,cr,cb;
    unsigned char re,gr,bl;

    int mean_cr = 0;
    int mean_cb = 0;

    /* Space for new image */
    dest->data=(unsigned char*)calloc(image->size*3,sizeof(unsigned char));

    dest->w= image->dx;
    dest->h= image->dy;
    dest->byte_per_line=dest->w*3;

    if(image->format == ascii)
        k6300_ascii_to_binary(image);

    if (full_range) {

        for (y = 0; y < image->dy; y += 2) {
            for (x = 0; x < image->dx/2; x++) {
                i = image->data[x+y*image->dx];
                stat[i]++;
                if (i < stat_min) stat_min = i;
                if (i > stat_max) stat_max = i;
            }
        }

        /* Adjust the intensity range... */

        range = (double)(stat_max - stat_min);

        if (range) {
            range = 255.9 / range;
            for (y = 0; y < image->dy; y++) {
                for (x = 0; x < image->dx; x++) {
                    int i = image->data[x+y*image->dx] - stat_min;
                    i = (int)(i * range);
                    if (i < 0) i = 0;
                    if (i > 255) i = 255;
                    image->data[x+y*image->dx] = i;
                }
            }
        }
    }

#if 0
    /* Simple 4x4 average algorithm.
     * created image effective resolution is 4 times smaller.
     */
    for (y = 0; y < image->dy; y += 2) {
        for (x = 0; x < image->dx; x += 2) {

            r0 = image->data[(x+1)+(y+0)*image->dx] * image->wb_r;
            b0 = image->data[(x+0)+(y+1)*image->dx] * image->wb_b;
            g0 = image->data[(x+0)+(y+0)*image->dx];
            g1 = image->data[(x+1)+(y+1)*image->dx];

            yy = (r0*wr + b0*wb + g0*wg + g1*wg) / (wb+wr+2*wg);
            cr = r0 - yy;
            cb = b0 - yy;

            mean_cr += cr;
            mean_cb += cb;

            r0 = yy + cr*4;  if (r0 < 0) r0 = 0; if (r0 > 255) r0 = 255;
            b0 = yy + cb*4;  if (b0 < 0) b0 = 0; if (b0 > 255) b0 = 255;
            g0 = yy;	     if (g0 < 0) g0 = 0; if (g0 > 255) g0 = 255;

            re= (r0 & 0x000000FF);
            gr= (g0 & 0x000000FF);
            bl= (b0 & 0x000000FF);


            dest->data[y*dest->byte_per_line+x*3]=re;
            dest->data[y*dest->byte_per_line+x*3+1]=gr;
            dest->data[y*dest->byte_per_line+x*3+2]=bl;

            dest->data[y*dest->byte_per_line+(x+1)*3]=re;
            dest->data[y*dest->byte_per_line+(x+1)*3+1]=gr;
            dest->data[y*dest->byte_per_line+(x+1)*3+2]=bl;

            dest->data[(y+1)*dest->byte_per_line+x*3]=re;
            dest->data[(y+1)*dest->byte_per_line+x*3+1]=gr;
            dest->data[(y+1)*dest->byte_per_line+x*3+2]=bl;

            dest->data[(y+1)*dest->byte_per_line+(x+1)*3]=re;
            dest->data[(y+1)*dest->byte_per_line+(x+1)*3+1]=gr;
            dest->data[(y+1)*dest->byte_per_line+(x+1)*3+2]=bl;

        }
    }

#else
    /* Using the four closest neighbors to compute RGB values for
     * each pixel
     */

    /* Space for image processing */
    data_copy=(unsigned char**)calloc(image->dx,sizeof(unsigned char*));
    for(i=0; i<image->dx;i++)
        data_copy[i]=(unsigned char*)calloc(image->dy,sizeof(unsigned char));


    for (y = 0; y < image->dy; y += 1) {
        for (x = 0; x < image->dx; x += 1) {
            data_copy[x][y]=image->data[x+y*image->dx];
        }
    }

    for (y = 1; y < image->dy-1; y += 2) {
        for (x = 1; x < image->dx-1; x += 2) {
            g[0]=data_copy[x-1][y-1];
            g[1]=data_copy[x][y];
            g[2]=data_copy[x+1][y-1];
            g[3]=data_copy[x+2][y];
            g[4]=data_copy[x-1][y+1];
            g[5]=data_copy[x][y+2];
            g[6]=data_copy[x+1][y+1];
            g[7]=data_copy[x+2][y+2];

            for(i=0;i<8;i++)
                g[i]=g[i]*wg;

            r[0]=data_copy[x][y-1]*image->wb_r;
            r[1]=data_copy[x+2][y-1]*image->wb_r;
            r[2]=data_copy[x][y+1]*image->wb_r;
            r[3]=data_copy[x+2][y+1]*image->wb_r;

            for(i=0;i<4;i++)
                r[i]=r[i]*wr;

            b[0]=data_copy[x-1][y]*image->wb_b;
            b[1]=data_copy[x+1][y]*image->wb_b;
            b[2]=data_copy[x-1][y+2]*image->wb_b;
            b[3]=data_copy[x+1][y+2]*image->wb_b;

            for(i=0;i<4;i++)
                b[i]=b[i]*wb;


            dest->data[y*dest->byte_per_line+x*3]=
                (r[0]+r[2])/2;
            dest->data[y*dest->byte_per_line+x*3+1]=
                (g[1]+g[0]+g[2]+g[4]+g[6])/5;
            dest->data[y*dest->byte_per_line+x*3+2]=
                (b[0]+b[1])/2;

            dest->data[y*dest->byte_per_line+(x+1)*3]=
                (r[0]+r[1]+r[2]+r[3])/4;
            dest->data[y*dest->byte_per_line+(x+1)*3+1]=
                (g[1]+g[2]+g[3]+g[6])/4;
            dest->data[y*dest->byte_per_line+(x+1)*3+2]=
                b[1];

            dest->data[(y+1)*dest->byte_per_line+x*3]=
                r[2];
            dest->data[(y+1)*dest->byte_per_line+x*3+1]=
                (g[1]+g[4]+g[6]+g[5])/4;
            dest->data[(y+1)*dest->byte_per_line+x*3+2]=
                (b[0]+b[1]+b[2]+b[3])/4;

            dest->data[(y+1)*dest->byte_per_line+(x+1)*3]=
                (r[2]+r[3])/2;
            dest->data[(y+1)*dest->byte_per_line+(x+1)*3+1]=
                (g[1]+g[3]+g[6]+g[5]+g[7])/5;
            dest->data[(y+1)*dest->byte_per_line+(x+1)*3+2]=
                (b[1]+b[3])/2;
        }
    }
#endif
}

/********************************************************************************/
void k6300_white_balance(K6300_Image * image)
{
    int mean_r = 0;
    int mean_g = 0;
    int mean_b = 0;
    int x,y;
    int r0,b0,g0,g1;

    for (y = 0; y < image->dy; y += 2) {
        for (x = 0; x < image->dx; x += 2) {

            r0 = image->data[(x+0)+(y+0)*image->dx];
            b0 = image->data[(x+1)+(y+1)*image->dx];
            g0 = image->data[(x+1)+(y+0)*image->dx];
            g1 = image->data[(x+0)+(y+1)*image->dx];

            mean_r += r0;
            mean_g += (g0+g1) / 2;
            mean_b += b0;
        }
    }

    mean_r /= (image->dx*image->dy/4);
    mean_g /= (image->dx*image->dy/4);
    mean_b /= (image->dx*image->dy/4);

    image->wb_b = ((float) (mean_g)) / mean_b;
    image->wb_r = ((float) (mean_g)) / mean_r;
}
